from typing import Type


class AntiRefMeta(type):
    def __str__(self):
        return self.__name__


class AntiForwardRef(metaclass=AntiRefMeta):
    """
    i.e:
    Union["someString"] would return a ForwardRef, this class is a simple hack
    to just return a type contains the name.
    """

    name: str


def anti_forward_ref(name: str) -> Type[AntiForwardRef]:
    return type(name, (AntiForwardRef,), {"name": name})



from jinja2 import Environment, PackageLoader, select_autoescape

from qtgql.codegen.objecttype import GqlType

env = Environment(loader=PackageLoader("qtgql.codegen.py"), autoescape=select_autoescape())

SchemaTemplate = env.get_template("schema.jinja.py")


def py_template(types: list[GqlType]) -> str:
    return SchemaTemplate.render(types=types)



from pathlib import Path
from typing import Type

import requests  # type: ignore
from attrs import define

from qtgql.codegen.introspection import SchemaEvaluator, introspection_query


@define
class QtGqlConfig:
    url: str
    output: Path
    """full path for the generated output file."""
    evaluator: Type[SchemaEvaluator] = SchemaEvaluator

    def fetch(self) -> None:
        res = requests.post(self.url, json={"query": introspection_query})
        introspected = res.json()["data"]
        self.evaluator.from_dict(introspected).dump(self.output)



import enum
from functools import cached_property
from typing import List, Optional, Union

from attrs import define

from qtgql.codegen.py.bases import _BaseQGraphQLObject
from qtgql.codegen.utils import AntiForwardRef
from qtgql.typingref import TypeHinter


class Kinds(enum.Enum):
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    ENUM = "ENUM"
    LIST = "LIST"
    UNION = "UNION"
    NON_NULL = "NON_NULL"

    def __getitem__(self, item):
        for f in Kinds:
            if f.name == item:
                return item
        raise KeyError(item, "is a wrong kind")


@define(slots=False)
class FieldProperty:
    name: str
    type: TypeHinter
    type_map: dict[str, "GqlType"]
    description: Optional[str] = ""

    @cached_property
    def deserializer(self) -> str:
        """If the field is optional this would be."""
        # every thing is possibly optional since you can query for only so or so fields.
        default = f"data.get('{self.name}', None)"
        t = self.type.type
        type_hinter = self.type
        if t is Optional:
            t = self.type.of_type[0].type
            type_hinter = self.type.of_type[0]

        if t in BuiltinScalars.values():
            return default
        if t in (list, List):
            inner = type_hinter.of_type[0].type
            # handle inner type
            assert issubclass(inner, AntiForwardRef)
            gql_type = self.type_map[inner.name]
            return f"cls.{_BaseQGraphQLObject.deserialize_list_of.__name__}(parent, data, {gql_type.model_name}, '{self.name}', {gql_type.name})"
        if t is Union:
            return (
                f"cls.{_BaseQGraphQLObject.deserialize_union.__name__}(parent, data, '{self.name}')"
            )
        # handle inner type
        assert issubclass(t, AntiForwardRef)
        if t.name in self.type_map.keys():
            return f"cls.{_BaseQGraphQLObject.deserialize_optional_child.__name__}(parent, data, {t.name}, '{self.name}')"
        raise NotImplementedError

    @cached_property
    def annotation(self) -> str:
        """
        :returns: Annotation of the field based on the type.
        """
        ret = self.type.as_annotation()
        # int, str, float etc...
        if ret in BuiltinScalars.values():
            return ret.__name__
        # handle Optional, Union, List etc...
        # removing redundant prefixes.
        return str(ret).replace("typing.", "").replace("qtgql.codegen.utils.", "")

    @cached_property
    def property_type(self) -> str:
        if (
            self.type.type in BuiltinScalars.values()
            or self.type.of_type in BuiltinScalars.values()
        ):
            return self.annotation
        return "QObject"

    @cached_property
    def setter_name(self) -> str:
        return self.name + "_setter"

    @cached_property
    def signal_name(self) -> str:
        return self.name + "Changed"

    @cached_property
    def private_name(self) -> str:
        return "_" + self.name


@define(slots=False)
class GqlType:
    kind: Kinds
    name: str
    fields: list["FieldProperty"]
    docstring: Optional[str] = ""

    @cached_property
    def model_name(self) -> str:
        return self.name + "Model"


BuiltinScalars: dict[str, type] = {
    "Int": int,
    "Float": float,
    "String": str,
    "ID": str,
    "Boolean": bool,
}



from __future__ import annotations
from PySide6.QtCore import Property, Signal, QObject
from typing import Optional, Union
from qtgql.codegen.py.bases import get_base_graphql_object, BaseModel

BaseObject = get_base_graphql_object()

{% for type in types %}
class {{ type.name }}(BaseObject):
    """{{  type.docstring  }}"""

    def __init__(self, parent: QObject = None, {% for f in type.fields %} {{f.name}}: {{f.annotation}} = None, {% endfor %}):
        super().__init__(parent){% for f in type.fields %}
        self.{{  f.private_name  }} = {{f.name}}{% endfor %}

    @classmethod
    def from_dict(cls, parent,  data: dict) -> {{type.name}}:
        return cls(
        parent=parent,{% for f in type.fields %}
        {{f.name}} = {{f.deserializer}},{% endfor %}
        )

    {% for f in type.fields %}
    {{ f.signal_name }} = Signal()

    def {{ f.setter_name }}(self, v: {{  f.annotation  }}) -> None:
        self.{{  f.private_name  }} = v
        self.{{  f.name  }}Changed.emit()

    @Property(type={{ f.property_type }}, fset={{ f.setter_name }}, notify={{f.signal_name}})
    def {{ f.name }}(self) -> {{ f.annotation }}:
        return self.{{  f.private_name  }}
    {% endfor %}
class {{ type.model_name }}(BaseModel):
    def __init__(self, data: list[{{  type.name  }}], parent: Optional[BaseObject] = None):
        super().__init__(data, parent)


{% endfor %}



from functools import cached_property
from pathlib import Path
from typing import Optional, Union

import graphql

from qtgql.codegen.objecttype import BuiltinScalars, FieldProperty, GqlType, Kinds
from qtgql.codegen.py.compiler import py_template
from qtgql.codegen.utils import anti_forward_ref
from qtgql.typingref import TypeHinter

introspection_query = graphql.get_introspection_query(descriptions=True)


class SchemaEvaluator:
    def __init__(self, introspection: dict, compiler=py_template):
        self.compiler = compiler
        self.introspection = introspection
        self._generated_types: dict[str, GqlType] = {
            scalar: GqlType(name=scalar, kind=Kinds.SCALAR, fields=[])
            for scalar in BuiltinScalars.keys()
        }
        self._evaluate()

    @cached_property
    def unions(self) -> list[dict]:
        return [
            t for t in self.introspection["__schema"]["types"] if Kinds[t["kind"]] == Kinds.UNION
        ]

    def get_possible_types_for_union(self, name: str) -> list[dict]:
        for union in self.unions:
            if union["name"] == name:
                return union["possibleTypes"]
        raise ValueError(f"Union for {name} was not found")

    def evaluate_field_type(self, t: dict) -> TypeHinter:
        kind = Kinds[t["kind"]]

        is_optional = True
        if kind == Kinds.NON_NULL:
            # move to inner, this is not optional
            t = t["ofType"]
            is_optional = False

        of_type = t["ofType"]
        name = t["name"]
        kind = Kinds[t["kind"]]
        ret: Optional[TypeHinter] = None

        def optional_maybe(inner: TypeHinter) -> TypeHinter:
            return TypeHinter(type=Optional, of_type=(inner,)) if is_optional else inner

        if kind == Kinds.LIST:
            ret = TypeHinter(type=list, of_type=(self.evaluate_field_type(of_type),))
        elif kind == Kinds.SCALAR:
            ret = TypeHinter(type=BuiltinScalars[name])
        elif kind == Kinds.OBJECT:
            ret = TypeHinter(type=anti_forward_ref(name))
        elif kind == Kinds.UNION:
            ret = TypeHinter(
                type=Union,
                of_type=tuple(
                    TypeHinter(type=anti_forward_ref(possible["name"]))
                    for possible in self.get_possible_types_for_union(name)
                ),
            )
        if ret:
            return optional_maybe(ret)

        raise NotImplementedError(f"kind {kind} not supported yet")

    def evaluate_field(self, field: dict) -> FieldProperty:
        """we don't really know what is the field type just it's name."""
        return FieldProperty(
            type=self.evaluate_field_type(field["type"]),
            name=field["name"],
            type_map=self._generated_types,
            description=field["description"],
        )

    def evaluate_object_type(self, type_: dict) -> Optional[GqlType]:
        # scalars are swallowed here.
        t_name: str = type_["name"]
        if t_name.startswith("__"):
            return None
        if evaluated := self._generated_types.get(t_name, None):
            return evaluated

        concrete = GqlType(
            kind=Kinds.OBJECT,
            name=t_name,
            docstring=type_["description"],
            fields=[self.evaluate_field(f) for f in type_["fields"]],
        )
        self._generated_types[t_name] = concrete
        return concrete

    def _evaluate(self) -> None:
        types = self.introspection["__schema"]["types"]
        for tp in types:
            kind = tp["kind"]
            if kind == Kinds.SCALAR.name:
                continue
            if kind == Kinds.OBJECT.name:
                self.evaluate_object_type(tp)

    def generate(self) -> str:
        """
        :return: The generated schema module as a string.
        """
        return self.compiler(
            [t for name, t in self._generated_types.items() if name not in BuiltinScalars.keys()]
        )

    @classmethod
    def from_dict(cls, introspection: dict) -> "SchemaEvaluator":
        evaluator = SchemaEvaluator(introspection)
        return evaluator

    def dump(self, file: Path):
        """
        :param file: Path to the file the codegen would dump to.
        """
        with open(file, "w") as fh:
            fh.write(self.generate())
